//=============================================================================
// Copyright  Micromega Corporation 2005-2008. All rights reserved.
//
// @file	GPSdemo1.c
// @target	Atmel WinAVR C compiler
//
// This demo program reads latitude and longitude values from a GPS receiver
// and displays them in the terminal window. The 4800 baud serial output from
// the GPS receiver is connected TO the SERIN pin of the uM-FPU V3.1 chip.
// The sentence count and sentence type is displayed for each record, and the
// longitude and latitude values are displayed for the GPRMC sentence.
// The sentence types will vary depending on the GPS receiver used.
// The following example shows output from a Garmin receiver:
//
//    GPSdemo1
//    uM-FPU V3.1.0
//
//    1:GPGLL
//    2:GPBOD
//    3:GPVTG
//    4:PGRME
//    5:PGRMZ
//    6:PGRMM
//    7:HCHDG
//    8:GPRTE
//    9:GPRMC: Lat = 7622.4609, Long = 7622.4609
//    (etc.)
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	October 15, 2007
//	- original version
//=============================================================================

#include <avr/io.h>
#include <stdio.h>
#include "uart.h"
#include "fpu.h"

FILE uart_stream = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);

//------------------ local prototypes -----------------------------------------

//------------------ uM-FPU register definitions ------------------------------

#define Latitude     10   		// latitude value
#define Longitude    11   		// longitude value

#define GPS_OK       0xFF80		// GPS sentence OK status

//------------------ main -----------------------------------------------------

int main(void) {
	int cnt;
	char status;

	// initialize the UART and set as stdout
	uart_init();
	stdout = &uart_stream;

	// print program name
	printf("\r\n\r\nGPSdemo1\r\n");

	// Reset the uM-FPU and print version.
	if (fpu_reset() == SYNC_CHAR) {
		print_version();
		print_CRLF();
	}
	else {
		printf("uM-FPU not detected");
		return(0);
	}

	// set FPU baud rate to 4800, enable NMEA input
	fpu_write3(SEROUT, 0, 5);
	fpu_write2(SERIN, 4);

	for (cnt = 0; cnt < 100; cnt++) {
		printf("\r\n%i:", cnt);

		fpu_write2(SERIN, 6);
		status = fpu_readStatus();

		if (status == GPS_OK) {
			print_fpuString(READSEL);

			fpu_write(STRCMP);
			fpu_writeString("GPRMC");
			status = fpu_readStatus();

			if (status & STATUS_ZERO) {
				fpu_write2(SELECTA, Latitude);
				fpu_write4(STRFIELD, 4, STRTOF, FSET0);
				fpu_write2(SELECTA, Longitude);
				fpu_write4(STRFIELD, 6, STRTOF, FSET0);
				fpu_write2(STRSET, 0);

				fpu_write2(SELECTA, Latitude);
				printf(" *** Lat = " );
				print_float(0);

				fpu_write2(SELECTA, Longitude);
				printf(", Long = ");
				print_float(0);
			}
		}
		else {
			printf(" *** data error ***");
		}
	}

	printf("\r\nDone.\r\n");
	return 0;
}
