//=============================================================================
// Copyright  Micromega Corporation 2005-2008. All rights reserved.
//
// @file	demo2.c
// @target	Atmel WinAVR C compiler
//
// This program demonstrates how to use the uM-FPU V3.1 floating point coprocessor
// connected to the Atmel AVR over an SPI interface. It takes temperature readings
// from a DS1620 digital thermometer, converts them to floating point and displays
// them in degrees Celsius and degrees Fahrenheit, and calculates the minimum and
// maximum temperatures..
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	February 4, 2008
//	- added delay before data sample in read_DS1620
//	October 14, 2007
//	- original version
//=============================================================================

#include <avr/io.h>
#include <stdio.h>
#include <util/delay.h>
#include "uart.h"
#include "fpu.h"

FILE uart_stream = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);

#define ddr_set(bit) (_SFR_BYTE(DDRC) = _SFR_BYTE(DDRC) | _BV(bit))
#define ddr_clr(bit) (_SFR_BYTE(DDRC) = _SFR_BYTE(DDRC) & ~_BV(bit))
#define bit_set(bit) (_SFR_BYTE(PORTC) = _SFR_BYTE(PORTC) | _BV(bit))
#define bit_clr(bit) (_SFR_BYTE(PORTC) = _SFR_BYTE(PORTC) & ~_BV(bit))
#define bit_read(bit) (_SFR_BYTE(PINC) & _BV(bit))

#define DS_RST	0
#define DS_CLK	1
#define DS_DATA	2

//------------------ local prototypes -----------------------------------------

void readTemp();
void printTempC(int reg);
void printTempF(int reg);
void init_DS1620(void);
void write_DS1620(char bval);
int read_DS1620(void);
void delay(int msec);

//-------------------- uM-FPU register definitions ----------------------------

#define DegC		1					// degrees Celsius
#define DegC_Min	2					// minimum temperature
#define DegC_Max	3					// maximum temperature
#define DegF		4					// degrees Fahrenheit
#define F1_8		5					// constant 1.8

//------------------ main -----------------------------------------------------

int main(void) {
			
	uart_init();					// initialize the UART
	stdout = &uart_stream;			// set stdout to UART

	printf("\r\n\r\nDemo2\r\n");		// program name

	ddr_set(DS_RST);
	ddr_set(DS_CLK);
	ddr_set(DS_DATA);

	if (fpu_reset() == SYNC_CHAR) {	// reset FPU and check synchronization
		print_version();			// print version
		print_CRLF();
	}
	else {
		printf("uM-FPU not detected");
		return(0);
	}

	init_DS1620();

 	// store constant value (1.8)
	fpu_write3(SELECTA, F1_8, ATOF);
	fpu_writeString("1.8");
	fpu_write(FSET0);

 	// set initial min/max values
    fpu_write4(SELECTA, DegC_Min, FSETI, 127);
    fpu_write4(SELECTA, DegC_Max, FSETI, -127);

    printf("\r\n              Min   Current   Max");
    printf("\r\n             -----  -------  -----");

    while (1) {

       // get temperature
      readTemp();

      // calculate minimum and maximum
      fpu_write4(SELECTA, DegC_Min, FMIN, DegC);
      fpu_write4(SELECTA, DegC_Max, FMAX, DegC);

     // display temperatures in Celsius
      printf("\r\nCelsius   ");
      printTempC(DegC_Min);
      printTempC(DegC);
      printTempC(DegC_Max);

      // display temperatures in Fahrenheit
      printf("\r\nFahrenheit");
      printTempF(DegC_Min);
      printTempF(DegC);
      printTempF(DegC_Max);

      // delay 2 seconds, then get the next reading
	  print_CRLF();
      delay(2000);
    }
}

//--------------------readTemp ---------------_--------------------------------

void readTemp() {
    // read raw temperature, send to uM-FPU, convert to floating point
    // and divide by 2 to get degrees Celsius
    fpu_write3(SELECTA, DegC, LOADWORD);
    fpu_writeWord(read_DS1620());
    fpu_write3(FSET0, FDIVI, 2);
}

//-------------------- printTempC ---------------------------------------------


void printTempC(int reg) {
    // print register value in 8.1 format

    fpu_write2(SELECTA, reg);
	print_float(81);
  }

//-------------------- printTempF ---------------------------------------------


void printTempF(int reg) {

    // degF = degC * 1.8 + 32
    fpu_write4(SELECTA, DegF, FSET, reg);
	fpu_write4(FMUL, F1_8, FADDI, 32);
	print_float(81);
  }

//-------------------- init_DS1620 --------------------------------------------

void init_DS1620(void) {

	bit_clr(DS_RST);
	bit_set(DS_CLK);
	delay(100);

	bit_set(DS_RST);
	write_DS1620(0x0C);
	write_DS1620(0x02);
	bit_clr(DS_RST);
	delay(100);

	bit_set(DS_RST);
	write_DS1620(0xEE);
	bit_clr(DS_RST);
	delay(1000);
}

//-------------------- write_DS1620 -------------------------------------------

void write_DS1620(char bval) {
	int i;

	for (i = 0; i < 8; i++) {
		bit_clr(DS_CLK);
		if (bval & 0x01)
			bit_set(DS_DATA);
		else
			bit_clr(DS_DATA);
		bval >>= 1;
		bit_set(DS_CLK);
	}
}

//-------------------- read_DS1620 --------------------------------------------

int read_DS1620(void) {
	int tmp, i;

	bit_set(DS_RST);
	write_DS1620(0xAA);

	ddr_clr(DS_DATA);
	tmp = 0;
	for (i = 0; i < 9; i++) {
		bit_clr(DS_CLK);
		_delay_us(1);
		tmp >>= 1;
		if (bit_read(DS_DATA)) tmp |= 0x0100;
		bit_set(DS_CLK);
	}
	bit_clr(DS_CLK);
	ddr_set(DS_DATA);

	bit_clr(DS_RST);
	if (tmp & 0x0100) tmp |= 0xFF00;
	return tmp;
}

//-------------------- delay --------------------------------------------------

void delay(int msec) {
	int i;

	for (i = 0; i < msec; i += 10) {
		_delay_ms(10);
	}
}
		
