//=============================================================================
// Copyright  Micromega Corporation 2005-2008. All rights reserved.
//
// @file	fpu_i2c.S
// @target	Atmel WinAVR C compiler
//
// Support routines for using the uM-FPU V3.1 floating point coprocessor with
// the Atmel AVR microcontroller over an I2C interface.
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	February 1, 2008
//	- added fpu_i2c.h
//	January 30, 2008
//	- original version
//=============================================================================

#include "gasava.inc"
#include "macros.inc"
#include "fpu_i2c.h"
#include "fops.h"

#define READ_DELAY	(4 * F_FPU / 40000)		// 400 usec ;*****temporary

	TEXT_SEG(fplib, fpu)

//------------------- fpu_reset -----------------------------------------------
// char fpu_reset(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_reset)

	ldi		r30, I2C_ADDR					; set I2C slave address
	sts		TWAR, r30

	ldi		r30, TWBR_VALUE					; set I2C bit rate
	sts		TWBR, r30

	rcall	i2c_start						; send start condition
	ldi		r30, I2C_ADDR					; send uM-FPU I2C device write address
	rcall	i2c_write
	ldi		r30, 1							; select I2C reset register
	rcall	i2c_write
	ldi		r30, 0							; reset the FPU
	rcall	i2c_write
	rcall	i2c_stop
					
	ldi		r30, RESET_DELAY & 0xFF			; delay for 10 msec
	ldi		r31, RESET_DELAY >> 8

resetLoop2:
	sbiw	r30, 1
	brne	resetLoop2	

	rcall	fpu_startWrite					; send SYNC command
	ldi		r30, SYNC
	rcall	i2c_write
	rjmp	fpu_read						; return sync reply

//------------------- fpu_wait ------------------------------------------------
// void fpu_wait(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_wait)
	push	rP1								; save rP1 register

fpu_wait1:
	rcall	fpu_startWrite					; start transmission

fpu_wait2:
	rcall	fpu_read						; wait for ready status
	cpi		rP1, 0
	brne	fpu_wait1

	pop		rP1								; restore register and exit
	ret

//------------------- fpu_write -----------------------------------------------
// void fpu_write(char bval1);
// void fpu_write2(char bval1, bval2);
// void fpu_write3(char bval1, bval2, bval3);
// void fpu_write4(char bval1, bval2, bval3, bval4);
//-----------------------------------------------------------------------------

GLOBAL(fpu_write4)

	rcall	fpu_startWrite					; start write transmission
	mov		r30, rP1						; write four bytes
	rcall	i2c_write
	mov		r30, rP3
	rcall	i2c_write
	mov		r30, rP5
	rcall	i2c_write
	mov		r30, rP7
	rjmp	writeByte

GLOBAL(fpu_write3)

	rcall	fpu_startWrite					; start write transmission
	mov		r30, rP1						; write three bytes
	rcall	i2c_write
	mov		r30, rP3
	rcall	i2c_write
	mov		r30, rP5
	rjmp	writeByte

GLOBAL(fpu_write2)

	rcall	fpu_startWrite					; start write transmission
	mov		r30, rP1						; write two bytes
	rcall	i2c_write
	mov		r30, rP3
	rjmp	writeByte

GLOBAL(fpu_write)

	rcall	fpu_startWrite					; start write transmission
	mov		r30, rP1						; write one byte

writeByte:
	rcall	i2c_write						; write byte
	rjmp	i2c_stop						; stop transmission

//------------------- fpu_writeWord -------------------------------------------
// void fpu_writeWord(int i);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeWord)

	rcall	fpu_startWrite					; start write transmission
	mov		r30, rP0						; write 16-bit value
	rcall	i2c_write
	mov		r30, rP1
	rjmp	writeByte

//------------------- fpu_writeFloat, fpu_writeLong ---------------------------
// void fpu_writeLong(long lval);
// void fpu_writeFloat(float fval);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeFloat)
GLOBAL(fpu_writeLong)

	rcall	fpu_startWrite					; start write transmission

write32:
	mov		r30, rP0						; write 32-bit value
	rcall	i2c_write
	mov		r30, rP1
	rcall	i2c_write
	mov		r30, rP2
	rcall	i2c_write
	mov		r30, rP3
	rjmp	writeByte

//------------------- fpu_writeString -----------------------------------------
// void fpu_writeString(char *s);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeString)

	rcall	fpu_startWrite					; start write transmission

	movw	r26, rP1						; load X register

writeString2:
	ld		r30, X							; get next character from string
	rcall	i2c_write						; write character

	ld		r30, X+							; repeat until zero terminator
	cpi		r30, 0
	brne	writeString2

	rjmp	i2c_stop						; stop transmission

//------------------- fpu_writeFloatReg, fpu_writeLongReg ---------------------
// void fpu_writeFloatReg(float fval, char reg);
// void fpu_writeLongReg(long lval, char reg);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeFloatReg)

	ldi		r31, FWRITE0					; write float
	rjmp	writeReg

GLOBAL(fpu_writeLongReg)

	ldi		r31, LWRITE0					; write long

writeReg:
	rcall	fpu_startWrite					; start write transmission
	mov		r30, r31

	cpi		rP5, 0							; check for reg 0
	breq	writeReg2

	subi	r30, 2							; check for reg X
	cpi		rP5, -1
	breq	writeReg2

	inc		r30								; check for reg A
	cpi		rP5, -2
	breq	writeReg2

	subi	r30, 2							; write to reg N
	rcall	i2c_write
	mov		r30, rP5

writeReg2:
	rcall	i2c_write						; send instruction
	rjmp	write32

//------------------- fpu_wrblk -----------------------------------------------
// void fpu_wrblk(char cnt, float *ptr);
//-----------------------------------------------------------------------------

GLOBAL(fpu_wrblk)

	rcall	fpu_wait						; wait until FPU is ready	
	rcall	fpu_startWrite					; start write transmission

	ldi		r30, WRBLK						; send WRBLK instruction
	rcall	i2c_write
	mov		r30, rP1						; send 32-bit word count
	andi	r30, 0x7F
	ori		r30, 0x80						; reverse the byte order
	rcall	i2c_write

	movw	r26, rP3						; load X register

wrblk2:
	ldi		rP0, 4							; get byte count

wrblk3:
	ld		r30, X+							; send next byte
	rcall	i2c_write
	dec		rP0
	brne	wrblk3							; repeat for byte count

	dec		rP1								; repeat for long word count
	brne	wrblk2

	rjmp	i2c_stop						; stop transmission

//------------------- fpu_read ------------------------------------------------
// char fpu_read(void);
//-----------------------------------------------------------------------------
				
GLOBAL(fpu_read)

	rcall	fpu_startRead					; start read transmission

readByte:
	rcall	i2c_readNAK
	mov		rP1, r30
	rjmp	i2c_stop						; stop transmission

//------------------- fpu_readWord --------------------------------------------
// int fpu_readWord(void);
//-----------------------------------------------------------------------------
	
GLOBAL(fpu_readWord)

	rcall	fpu_startRead					; start read transmission

	rcall	i2c_readACK						; read high byte
	mov		rP0, r30
	rjmp	readByte						; read low byte and exit

//------------------- fpu_readFloat, fpu_readLong -----------------------------
// float fpu_readFloat(void);
// long fpu_readLong(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readFloat)
GLOBAL(fpu_readLong)

	rcall	fpu_startRead					; start read transmission

	rcall	i2c_readACK						; read 32-bit value
	mov		rP0, r30
	rcall	i2c_readACK
	mov		rP1, r30
	rcall	i2c_readACK
	mov		rP2, r30
	rcall	i2c_readNAK
	mov		rP3, r30

	rjmp	i2c_stop						; stop transmission

//------------------- fpu_readString ------------------------------------------
// void fpu_readString(char *s);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readString)

	rcall	fpu_startRead					; start read transmission

	movw	r26, rP1						; load X register

readString2:
	rcall	i2c_readACK						; read character
	st		X+, r30							; store character in string
	cpi		r30, 0
	brne	readString2						; repeat until zero terminator

	rcall	i2c_readNAK						; read extra byte to send NAK
	rjmp	i2c_stop						; stop transmission

//------------------- fpu_readFloatReg, fpu_readLongReg -----------------------
// float fpu_readFloatReg(char reg);
// long fpu_readLongReg(char reg);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readFloatReg)

	ldi		r31, FREAD0						; read float
	rjmp	readReg

GLOBAL(fpu_readLongReg)

	ldi		r31, LREAD0						; read long

readReg:
	rcall	fpu_wait						; wait until FPU is ready
	rcall	fpu_startWrite					; start write transmission
	mov		r30, r31

	cpi		rP1, 0							; check for reg 0
	breq	readReg2

	subi	r30, 2							; check for reg X
	cpi		rP1, -1
	breq	readReg2

	inc		r30								; check for reg A
	cpi		rP1, -2
	breq	readReg2

	subi	r30, 2							; write to reg N
	rcall	i2c_write
	mov		r30, rP1

readReg2:
	rcall	i2c_write						; send instruction
	rcall	i2c_stop						; stop transmission
		
	rjmp	fpu_readLong					; read 32-bit value and exit

//------------------- fpu_readStatus ------------------------------------------
// char fpu_readStatus(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readStatus)

	rcall	fpu_wait						; wait until FPU is ready
	rcall	fpu_startWrite					; start write transmission	

readStatus:
	ldi		r30, READSTATUS					; send READSTATUS instruction
	rcall	i2c_write
	rcall	i2c_stop						; stop transmission

	rjmp	fpu_read						; read status byte and exit

//------------------- fpu_rdblk -----------------------------------------------
// void fpu_rdblk(char cnt, float *ptr);
//-----------------------------------------------------------------------------

GLOBAL(fpu_rdblk)

	rcall	fpu_wait						; wait until FPU is ready
	rcall	fpu_startWrite					; start write transmission

	ldi		r30, RDBLK						; send RDBLK instruction
	rcall	i2c_write
	mov		r30, rP1						; send 32-bit word count
	andi	r30, 0x7F
	ori		r30, 0x80						; reverse the byte order
	rcall	i2c_write

	rcall	fpu_startRead					; start read transmission
	movw	r26, rP3						; load X register

rdblk2:
	ldi		rP0, 4							; get byte count

rdblk3:
	rcall	i2c_readACK						; get character from SPI
	st		X+, r30							; store character
	dec		rP0
	brne	rdblk3							; repeat for byte count

	dec		rP1								; repeat for long word count
	brne	rdblk2

	rcall	i2c_readNAK						; read extra byte to send NAK
	rjmp	i2c_stop						; stop transmission

	rjmp	readStatus						; (workaround for V3.1.1 bug 1)

//------------------- fpu_fcall -----------------------------------------------
// void fpu_fcall(char func);
//-----------------------------------------------------------------------------

GLOBAL(fpu_fcall)
	
	ldi		rP0, FCALL						; function call
	rjmp	fpu_writeWord

//------------------- fpu_readDelay -------------------------------------------
// void fpu_readDelay(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readDelay)

	ldi		r30, READ_DELAY & 0xFF			; delay for 500 usec
	ldi		r31, READ_DELAY >> 8			; temporary workaround

readDelay:
	sbiw	r30, 1
	brne	readDelay	

	ret

//------------------- _floatOp1, _floatOp2, _floatOp3 -------------------------
// floatOp1		single parameter, 1-byte opcode
// floatOp2		two parameter, 1-byte opcode
// floatOp3		two parameter, 2-byte opcode
//-----------------------------------------------------------------------------

GLOBAL(_floatOp1)
	rcall	_loadRegA						; one operand (1-byte instruction)
	rjmp	floatOp

GLOBAL(_floatOp2)
	rcall	_loadRegA						; two operands (1-byte instruction)
	rcall	_loadReg0
	rjmp	floatOp

GLOBAL(_floatOp3)							; two operands (2-byte instruction)
	rcall	_loadRegA
	rcall	_loadReg0

	mov		r30, r31						; send opcode, 0
	rcall	i2c_write
	clr		r30
	rjmp	floatOp_exit

floatOp:
	mov		r30, r31						; send opcode

floatOp_exit:
	rcall	i2c_write						; send opcode
	rjmp	_saveReg0						; read result and exit

//------------------- _loadRegA, _loadReg0, _saveReg0 -------------------------
// _loadRegA	LEFT, FWRITEA, rA3, rA2, rA1, rA0
// _loadReg0	FWRITE0, rB3, rB2, rB1, rB0
// _saveReg0	RIGHT, wait, FREAD0, rP0, rP1, rP2, rP3
//-----------------------------------------------------------------------------

GLOBAL(_loadRegA)

	rcall	fpu_startWrite					; start write transmission	
	ldi		r30, LEFT						; LEFT
	rcall	i2c_write

	ldi		r30, FWRITEA					; send FWRITEA, rA3, rA2, rA1, rA0
	rcall	i2c_write
	mov		r30, rP0						; write 32-bit value
	rcall	i2c_write
	mov		r30, rP1
	rcall	i2c_write
	mov		r30, rP2
	rcall	i2c_write
	mov		r30, rP3
	rjmp	i2c_write

GLOBAL(_loadReg0)
	
	ldi		r30, FWRITE0					; FWRITE0, rp4, rP5, rP6, rP7
	rcall	i2c_write
	mov		r30, rB3
	rcall	i2c_write
	mov		r30, rB2
	rcall	i2c_write
	mov		r30, rB1
	rcall	i2c_write
	mov		r30, rB0
	rjmp	i2c_write

GLOBAL(_saveReg0)

	ldi		r30, RIGHT						; RIGHT
	rcall	i2c_write
	rcall	i2c_stop

	rcall	fpu_wait						; wait for result	

	rcall	fpu_startWrite					; start write transmission
	ldi		r30, FREAD0						; send FREAD0 instruction
	rcall	i2c_write
	rjmp	fpu_readLong					; read 32-bit value and exit

//------------------- fpu_startWrite ------------------------------------------

fpu_startWrite:
	rcall	i2c_start						; send start condition
	ldi		r30, I2C_ADDR					; send uM-FPU I2C device write address
	rcall	i2c_write
	ldi		r30, 0							; select I2C data register
	rjmp	i2c_write

//------------------- fpu_startRead -------------------------------------------

fpu_startRead:
	rcall	i2c_start						; send repeated start condition
	ldi		r30, I2C_ADDR+1					; send uM-FPU I2C device read address
	rjmp	i2c_write

//------------------- i2c_start -----------------------------------------------

i2c_start:
	ldi		r30, (1<<TWINT) | (1<<TWSTA) | (1<<TWEN)
	sts		TWCR, r30						; send start condition

i2c_start2:
	lds		r30, TWCR						; wait for completion
	sbrs	r30, TWINT
	rjmp	i2c_start2
	ret

//------------------- i2c_write -----------------------------------------------

i2c_write:
	sts		TWDR, r30						; send data byte
	ldi		r30, (1<<TWINT) | (1<<TWEN)
	sts		TWCR, r30

i2c_write2:
	lds		r30, TWCR						; wait for completion
	sbrs	r30, TWINT
	rjmp	i2c_write2
	ret

//------------------- i2c_readNAK ---------------------------------------------

i2c_readNAK:
	ldi		r30, (1<<TWINT) | (1<<TWEN)
	sts		TWCR, r30						; set to NAK next byte received

i2c_readNAK2:
	lds		r30, TWCR						; wait for the next byte
	sbrs	r30, TWINT
	rjmp	i2c_readNAK2

	lds		r30, TWDR						; return the received byte
	ret

//------------------- i2c_readACK ---------------------------------------------

i2c_readACK:
	ldi		r30, (1<<TWINT) | (1<<TWEA) | (1<<TWEN)
	sts		TWCR, r30						; set to ACK the next byte received

i2c_readACK2:
	lds		r30, TWCR						; wait for the next byte
	sbrs	r30, TWINT
	rjmp	i2c_readACK2

	lds		r30, TWDR						; return the received byte

	cpi		r30, 0							; (workaround for V3.1.2 bug 2)
	brlt	i2c_readACK3					; check if bit 7 is set
	ret

i2c_readACK3:
	push	r30								; if bit 7 set, start a new read transfer
	rcall	i2c_stop
	rcall	fpu_startWrite
	rcall	fpu_startRead
	pop		r30
	ret

//------------------- i2c_stop ------------------------------------------------

i2c_stop:
	ldi		r30, (1<<TWINT) | (1<<TWSTO) | (1<<TWEN)
	sts		TWCR, r30						; send the stop condition
	ret


