//=============================================================================
// Copyright  Micromega Corporation 2005-2008. All rights reserved.
//
// @file	fpu_spi.S
// @target	Atmel WinAVR C compiler
//
// Support routines for using the uM-FPU V3.1 floating point coprocessor with
// the Atmel AVR microcontroller over an SPI interface.
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	February 1, 2008
//	- added support for ATmega164P/324P/644P, ATmega162
//	January 30, 2008
//	- renamed from fpu_hw.S to fpu_spi.S
//	- updated symbols to support ATmega48/88/168, ATmega64/128
//	October 31, 2007
//	- added fpu_wrblk and fpu_rdblk routines
//	September 12, 2007
//	- updated for uM-FPU V3.1
//	- added support ATmega8/16/32
//	November 22, 2006
//	- original version
//=============================================================================

#include "gasava.inc"
#include "macros.inc"
#include "fpu_spi.h"
#include "fops.h"

	TEXT_SEG(fplib, fpu)

//------------------- fpu_reset -----------------------------------------------
// char fpu_reset(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_reset)

	out		_SPCR, r1						; disable SPI

	cbi		SCK_PORT, SCK_BIT				; initialize SPI pins
	sbi		SCK_DDR, SCK_BIT

	sbi		MOSI_PORT, MOSI_BIT
	sbi		MOSI_DDR, MOSI_BIT

	sbi		MISO_DDR, MISO_BIT

	cbi		SS_PORT, SS_BIT					; *SS must be output
	sbi		SS_DDR, SS_BIT

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT			; enable the FPU chip select
	sbi		FPU_CS_DDR, FPU_CS_BIT
#endif

	ldi		r30, 80							; send reset pulses

resetLoop1:
	sbi		SCK_PORT, SCK_BIT
	cbi		SCK_PORT, SCK_BIT
	dec		r30
	brne	resetLoop1

	cbi		MOSI_PORT, MOSI_BIT
					
	ldi		r30, RESET_DELAY & 0xFF			; delay for 10 msec
	ldi		r31, RESET_DELAY >> 8

resetLoop2:
	sbiw	r30, 1
	brne	resetLoop2	
									
#if F_FPU <= 10000000
	sbi		_SPSR, SPI2X					; FOSC/2 or FOSC/4
#endif

	sbi		_SPCR, MSTR						; master mode
	sbi		_SPCR, SPE	 					; enable SPI
			
	ldi		r30, SYNC						; send SYNC instruction
	rjmp	readStatus						; return status byte

//------------------- fpu_wait ------------------------------------------------
// void fpu_wait(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_wait)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU

	ldi		r30, WAIT_DELAY+1				; wait for CS delay
waitDelay:
	dec		r30
	brne	waitDelay
#endif

wait2:
	sbic	MISO_PIN, MISO_BIT
	rjmp	wait2							; wait for uM-FPU

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_write -----------------------------------------------
// void fpu_write(char bval1);
// void fpu_write2(char bval1, bval2);
// void fpu_write3(char bval1, bval2, bval3);
// void fpu_write4(char bval1, bval2, bval3, bval4);
//-----------------------------------------------------------------------------

GLOBAL(fpu_write4)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	mov		r30, rP1
	rcall	spiWrite
	mov		r30, rP3
	rcall	spiWrite
	mov		r30, rP5
	rcall	spiWrite
	mov		r30, rP7
	rjmp	writeByte

GLOBAL(fpu_write3)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	mov		r30, rP1
	rcall	spiWrite
	mov		r30, rP3
	rcall	spiWrite
	mov		r30, rP5
	rjmp	writeByte

GLOBAL(fpu_write2)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	mov		r30, rP1
	rcall	spiWrite
	mov		r30, rP3
	rjmp	writeByte

GLOBAL(fpu_write)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	mov		r30, rP1

writeByte:
	rcall	spiWrite

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif
	ret


//------------------- fpu_writeWord -------------------------------------------
// void fpu_writeWord(int i);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeWord)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	mov		r30, rP0
	rcall	spiWrite
	mov		r30, rP1
	rjmp	writeByte

//------------------- fpu_writeFloat, fpu_writeLong ---------------------------
// void fpu_writeLong(long lval);
// void fpu_writeFloat(float fval);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeFloat)
GLOBAL(fpu_writeLong)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

write32:
	mov		r30, rP0						; send 32-bit value
	rcall	spiWrite
	mov		r30, rP1
	rcall	spiWrite
	mov		r30, rP2
	rcall	spiWrite
	mov		r30, rP3
	rjmp	writeByte

//------------------- fpu_writeString -----------------------------------------
// void fpu_writeString(char *s);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeString)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif
	
	movw	r26, rP1						; load X register

writeString2:
	ld		r30, X+							; get next character
	rcall	spiWrite						; send to SPI
	cpi		r30, 0
	brne	writeString2					; repeat until zero terminator

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_writeFloatReg, fpu_writeLongReg ---------------------
// void fpu_writeFloatReg(float fval, char reg);
// void fpu_writeLongReg(long lval, char reg);
//-----------------------------------------------------------------------------

GLOBAL(fpu_writeFloatReg)

	ldi		r30, FWRITE0					; write float
	rjmp	writeReg

GLOBAL(fpu_writeLongReg)

	ldi		r30, LWRITE0					; write long

writeReg:

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	cpi		rP5, 0							; check for reg 0
	breq	writeReg2

	subi	r30, 2							; check for reg X
	cpi		rP5, -1
	breq	writeReg2

	inc		r30								; check for reg A
	cpi		rP5, -2
	breq	writeReg2

	subi	r30, 2							; write to reg N
	rcall	spiWrite
	mov		r30, rP5

writeReg2:
	rcall	spiWrite						; send instruction		
	rjmp	write32

//------------------- fpu_wrblk -----------------------------------------------
// void fpu_wrblk(char cnt, float *ptr);
//-----------------------------------------------------------------------------

GLOBAL(fpu_wrblk)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

wrblk1:
	sbic	MISO_PIN, MISO_BIT				; wait for uM-FPU
	rjmp	wrblk1

	ldi		r30, WRBLK						; send WRBLK instruction
	rcall	spiWrite
	mov		r30, rP1
	andi	R30, 0x7F
	ori		R30, 0x80						; reverse the byte order
	rcall	spiWrite	

	movw	r26, rP3						; load X register

wrblk2:
	ldi		rP0, 4							; get byte count

wrblk3:
	ld		r30, X+							; get next byte
	rcall	spiWrite						; send to SPI
	dec		rP0
	brne	wrblk3							; repeat for byte count

	dec		rP1								; repeat for long word count
	brne	wrblk2

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_read ------------------------------------------------
// char fpu_read(void);
//-----------------------------------------------------------------------------
				
GLOBAL(fpu_read)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	rcall	fpu_readDelay					; wait for read setup delay

readByte:
	rcall	spiRead							; read byte
	mov		rP1, r30

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_readWord --------------------------------------------
// int fpu_readWord(void);
//-----------------------------------------------------------------------------
	
GLOBAL(fpu_readWord)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	rcall	fpu_readDelay					; wait for read setup delay

	rcall	spiRead							; read high byte
	mov		rP0, r30
	rjmp	readByte						; read low byte and exit

//------------------- fpu_readFloat, fpu_readLong -----------------------------
// float fpu_readFloat(void);
// long fpu_readLong(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readFloat)
GLOBAL(fpu_readLong)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

read32:
	rcall	fpu_readDelay					; wait for read setup delay

	rcall	spiRead							; read 32-bit value
	mov		rP0, r30
	rcall	spiRead
	mov		rP1, r30
	rcall	spiRead
	mov		rP2, r30
	rcall	spiRead
	mov		rP3, r30

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_readString ------------------------------------------
// void fpu_readString(char *s);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readString)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	rcall	fpu_readDelay					; wait for read setup delay	
	movw	r26, rP1						; load X register

readString2:
	rcall	spiRead							; read next character
	st		X+, r30							; store character
	cpi		r30, 0
	brne	readString2						; repeat until zero terminator

#ifdef FPU_CS_ENABLED
	sbi		FPU_CS_PORT, FPU_CS_BIT 		; disable FPU
#endif

	ret

//------------------- fpu_readFloatReg, fpu_readLongReg -----------------------
// float fpu_readFloatReg(char reg);
// long fpu_readLongReg(char reg);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readFloatReg)

	ldi		r30, FREAD0						; read float
	rjmp	readReg

GLOBAL(fpu_readLongReg)

	ldi		r30, LREAD0						; read long

readReg:

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

readReg_wait:
	sbic	MISO_PIN, MISO_BIT				; wait for uM-FPU
	rjmp	readReg_wait

	cpi		rP1, 0							; check for reg 0
	breq	readReg2

	subi	r30, 2							; check for reg X
	cpi		rP1, -1
	breq	readReg2

	inc		r30								; check for reg A
	cpi		rP1, -2
	breq	readReg2

	subi	r30, 2							; write to reg N
	rcall	spiWrite
	mov		r30, rP1

readReg2:
	rcall	spiWrite						; send instruction		
	rjmp	read32

//------------------- fpu_readStatus ------------------------------------------
// char fpu_readStatus(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readStatus)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

readStatus_wait:
	sbic	MISO_PIN, MISO_BIT				; wait for uM-FPU
	rjmp	readStatus_wait

	ldi		r30, READSTATUS					; send READSTATUS instruction

readStatus:
	rcall	spiWrite						; send status instruction
	rcall	fpu_readDelay					; wait for read setup delay
	rjmp	readByte						; read status byte and exit

//------------------- fpu_rdblk -----------------------------------------------
// void fpu_rdblk(char cnt, float *ptr);
//-----------------------------------------------------------------------------

GLOBAL(fpu_rdblk)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

rdblk1:
	sbic	MISO_PIN, MISO_BIT				; wait for uM-FPU
	rjmp	rdblk1

	ldi		r30, RDBLK						; send RDBLK instruction
	rcall	spiWrite
	mov		r30, rP1
	andi	R30, 0x7F
	ori		R30, 0x80						; reverse the byte order
	rcall	spiWrite	

	rcall	fpu_readDelay					; wait for read setup delay	

	movw	r26, rP3						; load X register

rdblk2:
	ldi		rP0, 4							; get byte count

rdblk3:
	rcall	spiRead							; read next byte
	st		X+, r30							; store the byte
	dec		rP0
	brne	rdblk3							; repeat for byte count

	dec		rP1								; repeat for long word count
	brne	rdblk2

	ldi		r30, READSTATUS					; workaround for pre-V3.1.2 bug
	rjmp	readStatus

//------------------- fpu_fcall -----------------------------------------------
// void fpu_fcall(char func);
//-----------------------------------------------------------------------------

GLOBAL(fpu_fcall)
	
	ldi		rP0, FCALL						; function call
	rjmp	fpu_writeWord

//------------------- fpu_readDelay -------------------------------------------
// void fpu_readDelay(void);
//-----------------------------------------------------------------------------

GLOBAL(fpu_readDelay)
	
#if READ_DELAY > 4
	ldi		r30, READ_DELAY - 4				; wait for read setup delay

readDelay:
	dec		r30
	brne	readDelay
#endif

	ret

//------------------- _floatOp1, _floatOp2, _floatOp3 -------------------------
// floatOp1		single parameter, 1-byte opcode
// floatOp2		two parameter, 1-byte opcode
// floatOp3		two parameter, 2-byte opcode
//-----------------------------------------------------------------------------

GLOBAL(_floatOp1)
	rcall	_loadRegA						; one operand (1-byte instruction)
	rjmp	floatOp

GLOBAL(_floatOp2)
	rcall	_loadRegA						; two operands (1-byte instruction)
	rcall	_loadReg0
	rjmp	floatOp

GLOBAL(_floatOp3)							; two operands (2-byte instruction)
	rcall	_loadRegA
	rcall	_loadReg0

	mov		r30, r31						; send opcode, 0
	rcall	spiWrite
	clr		r30
	rjmp	floatOp_exit

floatOp:
	mov		r30, r31						; send opcode

floatOp_exit:
	rcall	spiWrite						; send opcode
	rjmp	_saveReg0						; read result and exit

//------------------- _loadRegA, _loadReg0, _saveReg0 -------------------------
// _loadRegA	LEFT, FWRITEA, rA3, rA2, rA1, rA0
// _loadReg0	FWRITE0, rB3, rB2, rB1, rB0
// _saveReg0	RIGHT, wait, FREAD0, rP0, rP1, rP2, rP3
//-----------------------------------------------------------------------------

GLOBAL(_loadRegA)

#ifdef FPU_CS_ENABLED
	cbi		FPU_CS_PORT, FPU_CS_BIT 		; enable FPU
#endif

	ldi		r30, LEFT						; LEFT
	rcall	spiWrite

	ldi		r30, FWRITEA					; send FWRITEA, rA3, rA2, rA1, rA0
	rcall	spiWrite
	rjmp	write32

GLOBAL(_loadReg0)

	ldi		r30, FWRITE0					; FWRITE0, rp4, rP5, rP6, rP7
	rcall	spiWrite
	mov		r30, rB3
	rcall	spiWrite
	mov		r30, rB2
	rcall	spiWrite
	mov		r30, rB1
	rcall	spiWrite
	mov		r30, rB0
	rjmp	spiWrite

GLOBAL(_saveReg0)

	ldi		r30, RIGHT						; RIGHT
	rcall	spiWrite

saveReg0_wait:
	sbic	MISO_PIN, MISO_BIT				; wait for result
	rjmp	saveReg0_wait

	ldi		r30, FREAD0						; send FREAD0 instruction
	rcall	spiWrite
	rjmp	read32							; read value and exit

//------------------- spiWrite ------------------------------------------------
// write data byte to SPI buffer
//-----------------------------------------------------------------------------

spiWrite:

	out		_SPDR, r30						; write byte to SPI

spiWrite2:
	sbis	_SPSR, SPIF						; wait until byte has been sent
	rjmp	spiWrite2
	ret

//------------------- spiRead -------------------------------------------------
// read byte from SPI buffer
//-----------------------------------------------------------------------------

spiRead:

	out		_SPDR, r1						; send zero byte to SPI

spiRead2:
	sbis	_SPSR, SPIF						; wait until byte has been sent
	rjmp	spiRead2

	in		r30, _SPDR						; read byte from SPI
	ret

