//=============================================================================
// Copyright  Micromega Corporation 2005-2007. All rights reserved.
//
// @file	sensirion.c
// @target	Atmel WinAVR C compiler
//
// This program provides an example of using the uM-FPU V3.1 floating point
// coprocessor with the ARMmite/ARMexpress using an SPI interface. It converts
// temperature and humidity readings from a Sensirion SHT1x sensor.
//
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	February 4, 2008
//	- changed default pins for Sensirion SHT1x
//	October 15, 2007
//	- original version
//=============================================================================

#include <avr/io.h>
#include <stdio.h>
#include <util/delay.h>
#include "uart.h"
#include "fpu.h"

FILE uart_stream = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);

#define ddr_set(bit) (_SFR_BYTE(DDRC) = _SFR_BYTE(DDRC) | _BV(bit))
#define ddr_clr(bit) (_SFR_BYTE(DDRC) = _SFR_BYTE(DDRC) & ~_BV(bit))
#define bit_set(bit) (_SFR_BYTE(PORTC) = _SFR_BYTE(PORTC) | _BV(bit))
#define bit_clr(bit) (_SFR_BYTE(PORTC) = _SFR_BYTE(PORTC) & ~_BV(bit))
#define bit_read(bit) (_SFR_BYTE(PINC) & _BV(bit))


#define SHT_CLK		0					// Sensirion SHT1x clock
#define SHT_DATA	1					// Sensirion SHT1x data

//------------------ local prototypes -----------------------------------------

void SHT_Temperature(void);
void SHT_Humidity(void);
void SHT_Reset(void);
char SHT_Write(char b);
char SHT_Read(char ack);
int SHT_ReadWord(void);
void SHT_Wait(void);
void delay(int msec);

//------------------ uM-FPU register definitions ------------------------------

#define RawTemp		10					// raw temperature
#define TempC		11					// temperature - Celsius
#define TempF		12					// temperature - Fahrenheit
#define RawHumidity	13					// raw humidity
#define RhLinear	14					// relative humidity - linear
#define RhTrue		15					// relative humidity - compensated
#define H			16					// H value for dew point calculation
#define DewPoint	17					// dew point

#define C2			20					// constants for RhLinear calculation
#define C3			21
#define T1			22					// constants for RhTrue calculation
#define T2			23					// constant for DewPoint calculation 
#define	K1			24
#define K2			25
#define K3			26
#define K4			27

//------------------ main -----------------------------------------------------

int main(void) {

	// initialize the UART and set as stdout
	uart_init();
	stdout = &uart_stream;

	// print program name
	printf("\r\n\r\nSensirion Demo\r\n");

	// reset the uM-FPU and print version.
	if (fpu_reset() == SYNC_CHAR) {
		print_version();
		print_CRLF();
	}
	else {
		printf("uM-FPU not detected");
		return(0);
	}

	// reset the SHT-11x
	SHT_Reset();

	// store floating point constants
	fpu_writeFloatReg(0.0405, C2);
	fpu_writeFloatReg(-2.8e-6, C3);
	fpu_writeFloatReg(0.01, T1);
	fpu_writeFloatReg(0.00008, T2);
	fpu_writeFloatReg(0.4343, K1);
	fpu_writeFloatReg(17.62, K2);
	fpu_writeFloatReg(243.12, K3);

	while(1) {
		// read temperature
		//-----------------
		SHT_Temperature();
		printf("Raw Temperature:  ");
		print_float(60);
		print_CRLF();

		// TempC = -40.0 + 0.01 * rawTemp
		// TempC = RawTemp / 100.0 - 40.0
		//-------------------------------	
		fpu_write4(SELECTA, TempC, FSET, RawTemp);
		fpu_write4(FDIVI, 100, FSUBI, 40);
		printf("Degrees C:        ");
		print_float(62);
		print_CRLF();

		// TempF = TempC converted to Fahrenheit
		//--------------------------------------
		fpu_write4(SELECTA, TempF, FSET, TempC);
		fpu_write2(FCNV, 1);
		printf("Degrees F:        ");
		print_float(62);
		print_CRLF();

		// read humidity
		//--------------
		SHT_Humidity();
		printf("Raw Humidity:     ");
		print_float(60);
		print_CRLF();

		// RhLinear = -4.0 + 0.0405 * rawHumidity + (-2.8E-6) * rawHumidity^2
		// RhLinear = ((rawHumidity * C3) + C2) * rawHumidity) - 4
		//-------------------------------------------------------------------
		fpu_write4(SELECTA, RhLinear, FSET, RawHumidity);
		fpu_write2(FMUL, C3);
		fpu_write4(FADD, C2, FMUL, RawHumidity);
		fpu_write2(FSUBI, 2);	
		printf("Humidity (linear):");
		print_float(62);
		print_CRLF();

		// RhTrue = (TempC - 25.0) * (0.01 + 0.00008 * RawHumidity) + RhLinear
		// RhTrue = (TempC - 25.0) * (T2 * RawHumidity + T1) + RhLinear
		//--------------------------------------------------------------------	
		fpu_write4(SELECTA, RhTrue, FSET, TempC);
		fpu_write2(FSUBI, 25);
		fpu_write3(LEFT, FSET, T2);
		fpu_write4(FMUL, RawHumidity, FADD, T1);	
		fpu_write4(RIGHT, FMUL0, FADD, RhLinear);
		printf("Humidity (true):  ");
		print_float(62);
		print_CRLF();

		// H = (log10(RhTrue) - 2.0) / 0.4343 + (17.62 * TempC) / (243.12 + TempC)
		// H = (log10(RhTrue) - 2.0) / K1 + (K2 * TempC) / (K3 + TempC)
		//------------------------------------------------------------------------
		fpu_write4(SELECTA, H, FSET, RhTrue);
		fpu_write3(LOG10, FSUBI, 2);
		fpu_write2(FDIV, K1);	
		fpu_write3(LEFT, FSET, K2);
		fpu_write3(FMUL, TempC, LEFT);
		fpu_write4(FSET, K3, FADD, TempC);
		fpu_write4(RIGHT, FDIV0, RIGHT, FADD0);

		// DewPoint = 243.12 * H / (17.62 - H)
		// DewPoint = K3 * H / (K2 - H)
		//------------------------------------
		fpu_write4(SELECTA, DewPoint, FSET, K3);
		fpu_write2(FMUL, H);
		fpu_write3(LEFT, FSET, K2);
		fpu_write4(FSUB, H, RIGHT, FDIV0);
		printf("Dew Point:        ");
		print_float(62);
		print_CRLF();

		delay(2000);
		print_CRLF();
	}
}

//-------------------- Sensirion SHT1x/SHT7x routines --------------------------

void SHT_Temperature(void) {				// read temperature and send to FPU
	char i;
	char ack;
	int tmp;

	for (i = 0; i < 100; i++) {
		ack = SHT_Write(0x03);
		if (!ack) break;
		SHT_Reset();						// if NAK, reset the device and retry
	}
	SHT_Wait();
	tmp = SHT_ReadWord();
	fpu_write3(SELECTA, RawTemp, LOADWORD);
	fpu_writeWord(tmp);
	fpu_write(FSET0);
}


void SHT_Humidity(void)	{					// read humidity and send to uM-FPU
	char i;
	char ack;
	int tmp;

	for (i = 0; i < 100; i++) {
		ack = SHT_Write(0x05);
		if (!ack) break;
		SHT_Reset();						// if NAK, reset the device and retry
	}
	SHT_Wait();
	tmp = SHT_ReadWord();
	fpu_write3(SELECTA, RawHumidity, LOADWORD);
	fpu_writeWord(tmp);
	fpu_write(FSET0);
}

void SHT_Reset(void) {						// reset the SHTx1
	int i;

	bit_clr(SHT_CLK);
	ddr_set(SHT_CLK);
	ddr_clr(SHT_DATA);

	for (i = 0; i < 9; i++) {
		bit_clr(SHT_CLK);
		bit_set(SHT_CLK);
	}
	bit_clr(SHT_CLK);
}

char SHT_Write(char bval) {
	int i;
	char ack;

	bit_set(SHT_CLK);						// send start sequence

	bit_clr(SHT_DATA);
	ddr_set(SHT_DATA);

	bit_clr(SHT_CLK);
	bit_set(SHT_CLK);

	bit_set(SHT_DATA);

	for (i = 0; i < 8; i++) {				// send command
		bit_clr(SHT_CLK);
		if (bval & 0x80)
			bit_set(SHT_DATA);
		else
			bit_clr(SHT_DATA);
		bval <<= 1;
		bit_set(SHT_CLK);
	}
	bit_clr(SHT_CLK);

	ddr_clr(SHT_DATA);						// read ACK/NAK	
	bit_set(SHT_CLK);
	ack = bit_read(SHT_DATA);
	bit_clr(SHT_CLK);

	return ack;
}

char SHT_Read(char ack) {
	char i, tmp;

	tmp = 0;								// read byte
	for (i = 0; i < 8; i++) {
		bit_set(SHT_CLK);
		tmp <<= 1;
		if (bit_read(SHT_DATA)) tmp |= 1;
		bit_clr(SHT_CLK);
	}

	if (ack == 0)							// send ACK/NAK
		bit_clr(SHT_DATA);
	else
		bit_set(SHT_DATA);
	ddr_set(SHT_DATA);
	bit_set(SHT_CLK);
	bit_clr(SHT_CLK);	
	ddr_clr(SHT_DATA);

	return tmp;
}

int SHT_ReadWord(void) {					// read 16-bit value
	int tmp;

	tmp = SHT_Read(0) << 8;
	tmp |= SHT_Read(1) & 0xFF;
	return tmp;
}
  
void SHT_Wait(void) {						// wait for ready (with 250 msec timeout)
	int i;

	for (i = 0; i < 250; i++) {
		if (bit_read(SHT_DATA) == 0) break;
		_delay_ms(1);
	}
}

//-------------------- delay --------------------------------------------------

void delay(int msec) {
	int i;

	for (i = 0; i < msec; i += 10) {
		_delay_ms(10);
	}
}
		
