//=============================================================================
// Copyright  Micromega Corporation 2005-2008. All rights reserved.
//
// @file	uart.c
// @target	Atmel WinAVR C compiler
//
// Low-level UART routine.
//
// @author	Cam Thompson, Micromega Corporation
// @version
//	February 1, 2008
//	- added support for ATmega164P/324P/644P, ATmega162
//	- added getchar() function
//	- added handling of newline character
//	January 30, 2008
//	- added support ATmega48/88/168, ATmega64/128
//	August 31, 2007
//	- updated for uM-FPU V3.1
//	- added support ATmega8/16/32
//	November 15, 2006
//	- original version
//=============================================================================

#include <avr/io.h>
#include <stdio.h>

// CPU frequency
#ifndef F_CPU
  #warning "F_CPU set to 16000000."
  #define F_CPU		16000000
#endif

// to set the UART to port 1, add -DUART=1 to Project Custom Options,
// or uncomment the following definition.
// #define UART 1

#ifndef UART
  #define UART 0
#endif

#if defined(_AVR_IOM8_H_) || defined(_AVR_IOM16_H_) || defined(_AVR_IOM32_H_)
  // ATmega 8, ATmega16, ATmega32
  #define UCSRC_VALUE _BV(URSEL) | _BV(UCSZ1) | _BV(UCSZ0)
  #undef UART
  #define UART -1

#elif defined(_AVR_IOMX8_H_)
  // ATmega48, ATmega88, ATmeta168
  #define UCSRC_VALUE _BV(UCSZ01) | _BV(UCSZ00)
  #if (UART > 0)
    #define UART 0
  #endif

#elif defined(_AVR_IOM64_H_) || defined(_AVR_IOM128_H_) || defined(_AVR_IOMXX4_H_)
  // ATmega64, ATmega128, ATmega164P, ATmega324P, ATmega644P
  #define UCSRC_VALUE _BV(UCSZ01) | _BV(UCSZ00)
  #if (UART > 1)
    #define UART 0
  #endif

#elif defined(_AVR_IOM162_H_)
  // ATmega162
  #if (UART == 1)
    #define UCSRC_VALUE _BV(URSEL1) |_BV(UCSZ01) | _BV(UCSZ00)
  #else
    #define UART 0
    #define UCSRC_VALUE _BV(URSEL0) |_BV(UCSZ01) | _BV(UCSZ00)
  #endif

#else
  #error "uart.c symbols not defined for selected processor"
#endif

#if (UART == 0)
  #define _UCSRA 	UCSR0A
  #define _UCSRB 	UCSR0B
  #define _UCSRC 	UCSR0C 
  #define _UBRRL 	UBRR0L
  #define _UBRRH 	UBRR0H
  #define _UDR 		UDR0
  #define _RXEN		RXEN0
  #define _TXEN 	TXEN0
  #define _UDRE 	UDRE0
  #define _RXC		RXC0

#elif (UART == 1)
  #define _UCSRA 	UCSR1A
  #define _UCSRB 	UCSR1B
  #define _UCSRC 	UCSR1C 
  #define _UBRRL 	UBRR1L
  #define _UBRRH 	UBRR1H
  #define _UDR 	UDR1
  #define _RXEN	RXEN1
  #define _TXEN 	TXEN1
  #define _UDRE 	UDRE1
  #define _RXC		RXC1

#else
  #define _UCSRA	UCSRA
  #define _UCSRB	UCSRB
  #define _UCSRC	UCSRC 
  #define _UBRRL 	UBRRL
  #define _UBRRH 	UBRRH
  #define _UDR 		UDR
  #define _RXEN		RXEN
  #define _TXEN 	TXEN
  #define _UDRE 	UDRE
  #define _RXC		RXC
#endif

// UART baud rate
#ifndef BAUD_RATE
  #warning "BAUD_RATE set to 19200"
  #define BAUD_RATE	19200
#endif

#define UBRR_VALUE ((F_CPU/(BAUD_RATE*16UL))-1)

//-----------------------------------------------------------------------------

void uart_init(void) {

   	_UCSRB = 0; 				// disable UART while setting baud rate
    _UCSRA = 0;
    _UCSRC = UCSRC_VALUE;
	_UBRRL = UBRR_VALUE;
	_UBRRH = (UBRR_VALUE >> 8);
    _UCSRB = _BV(_RXEN) | _BV(_TXEN);
}

//-----------------------------------------------------------------------------

int uart_putchar(char c, FILE *stream) {

	if (c == '\n')
		uart_putchar('\r', stream);
	loop_until_bit_is_set(_UCSRA, _UDRE);
	_UDR = c;
	return 0;
}

//-----------------------------------------------------------------------------

int uart_getchar(FILE *stream) {
	unsigned char c;

	loop_until_bit_is_set(_UCSRA, _RXC);
	c = _UDR;
	uart_putchar(c, stream);	// echo the received character
	if (c == '\r') {
		c = '\n';
		uart_putchar(c, stream);
	}
	return c;
}
